#!/usr/bin/env python
'''
engine_mx
Created by Seria at 12/02/2019 3:45 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import os
import torch
from ..kit.utility import GPUtil

CPU = 0
GPU = 1


class Engine(object):
    '''
    Param:
    device: CPU or GPU
    available_gpus
    gpu_mem_fraction
    if_conserve
    least_mem
    '''
    def __init__(self, device=GPU, ngpu=1, least_mem=2048, avail_gpus=(), multi_piston=False):
        self.rank = int(os.environ.get('RANK', -1))
        self.device = device
        self.multi_piston = multi_piston
        # look for available gpu devices
        if self.device == GPU:
            if len(avail_gpus) == 0:
                gputil = GPUtil()
                gpus = gputil.available(ngpu, least_mem)
            else:
                gpus = avail_gpus
            if len(gpus) < ngpu:
                raise Exception('NEBULAE ERROR ⨷ no enough available gpu.')
            # convert gpu list to string
            str_gpus = ','.join([str(g[0]) for g in gpus])
            # set environment variable
            os.environ['CUDA_VISIBLE_DEVICES'] = str_gpus
            self.chip = [torch.device('cuda:%d'%i) for i in range(ngpu)]
            # setup multi-gpu environment
            if self.rank>=0:
                torch.backends.cudnn.benchmark = True
                torch.cuda.set_device(self.rank)
                torch.distributed.init_process_group(backend='nccl', init_method='env://')
            if self.rank<=0:
                print('+' + 20 * '-' + '+')
                print('| Reside in Devices: |')
                print('+' + 70 * '-' + '+')
                for g in gpus:
                    print('| \033[1;36mGPU {:<2d}\033[0m | {:<25s} | {:>5d} MiB free out of {:<5d} MiB |'.format(
                        g[0], g[1], g[3], g[2]
                    ))
                    print('+' + 70 * '-' + '+')
        elif self.device == CPU:
            assert self.rank<0
            print('+' + (24 * '-') + '+')
            print('| Reside in Devices: \033[1;36mCPU\033[0m |')
            print('+' + (24 * '-') + '+')
        else:
            raise KeyError('NEBULAE ERROR ⨷ given device should be either cpu or gpu.')

        ###################################
        ##          IT WORKS !           ##
        ###################################
        from ..astro import dock
        dock.coat = self.coat
        dock.shell = self.shell

    def coat(self, datum, as_const=True, sync=True):
        if not isinstance(datum, torch.Tensor): # numpy array
            if isinstance(datum, (int, float)) or datum.shape==(): # scalar
                datum = torch.tensor(datum)
            else:
                datum = torch.from_numpy(datum)

        if as_const:
            datum.requires_grad = False
        else:
            datum.requires_grad = True

        if self.device == GPU:
            if self.rank<0:
                return datum.cuda(non_blocking=not sync)
            else:
                return datum.to(self.chip[self.rank], non_blocking=not sync)
        elif self.device == CPU:
            return datum.cpu()
        else:
            raise KeyError('NEBULAE ERROR ⨷ given device should be either cpu or gpu.')

    def shell(self, datum, as_np=True, sync=False):
        datum = datum.detach()
        # add a sync point
        if sync and self.device == GPU:
            if self.rank < 0:
                torch.cuda.synchronize()
            else:
                torch.cuda.synchronize(self.chip[self.rank])
        if as_np:
            if datum.size == 1:
                datum = datum.cpu().numpy()[0]
            else:
                datum = datum.cpu().numpy()
        return datum


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import numpy as np
from scipy import special, signal
import random as rand
import logging
import math
import cv2
from PIL import Image, ImageFilter
from torchvision.transforms import *
F = functional
from collections import abc
from io import BytesIO
from ..kit import byte2arr
try:
    import av
    logging.getLogger('libav').setLevel(logging.FATAL)
except:
    print('NEBULAE WARNING ◘ that PyAV has not been installed properly results in failure of video augmentation.')


__all__ = ('Comburant', 'HWC2CHW', 'Random',
           'multiple',
           'NEAREST', 'LINEAR', 'CUBIC', 'HORIZONTAL', 'VERTICAL',
           'Pad', 'Resize',
           'Crop', 'Flip', 'Rotate',
           'Brighten', 'Contrast', 'Saturate', 'Hue', 'Blur', 'Sharpen',
           'Noise', 'Sinc',
           'JPEG', 'WebP', 'MPEG4', 'H264', 'VP9', 'AV1')


NUMPY = 0
PIL = 1

NEAREST = 10
LINEAR = 11
CUBIC = 12

CV_INTER = {NEAREST: cv2.INTER_NEAREST, LINEAR: cv2.INTER_LINEAR, CUBIC: cv2.INTER_CUBIC}
PIL_INTERP = {NEAREST: Image.NEAREST, LINEAR: Image.BILINEAR, CUBIC: Image.BICUBIC}

HORIZONTAL = 20
VERTICAL = 21

GAUSSIAN = 30
POISSON = 31




class Comburant(object):
    def __init__(self, *args, format=NUMPY, is_encoded=False, end=True):
        if format == NUMPY:
            pass
        elif format == PIL:
            pass
        else:
            raise Exception('NEBULAE ERROR ⨷ the image format is not defined or supported.')
        self.format = format
        if len(args) > 0 and isinstance(args[-1], HWC2CHW):
            ls_args = list(args[:-1])
            self.cvt_form = args[-1]
        else:
            ls_args = list(args)
        for a in args:
            a._format = format
        self.comburants = ls_args
        self.is_encoded = is_encoded
        self.end = end

    def _byte2arr(self, imgs):
        if isinstance(imgs, abc.Sequence):
            arr = []
            for i in imgs:
                arr.append(self._byte2arr(i))
        else:
            if self.is_encoded:
                arr = byte2arr(imgs, as_np=self.format==NUMPY)
            elif imgs.dtype != np.uint8:
                arr = (imgs * 255).astype(np.uint8)
            else:
                arr = imgs
        return arr

    def _post(self, src):
        if isinstance(src, abc.Sequence):
            dst = []
            for img in src:
                dst.append(self._post(img))
        else:
            if not isinstance(src, np.ndarray):
                src = np.array(src)
                dst = src.astype(np.float32) / 255.
            else:
                dst = src.astype(np.float32) / 255.
            if hasattr(self, 'cvt_form'):
                dst = self.cvt_form(dst)
        return dst

    def __call__(self, src):
        # >| is_encoded = True:  decode image bytes
        # >| is_encoded = False: convert to uint8 for numpy format
        if self.is_encoded or self.format == NUMPY:
            src = self._byte2arr(src)
        # >| go through comburants
        for cbr in self.comburants:
            src = cbr(src)
        if self.end:
            dst = self._post(src)
            return dst
        else:
            return src



class ABC(object):
    def __init__(self, pair_fn=None):
        self.pair_fn = pair_fn
        self._format = None

    def call(self, *args):
        raise NotImplementedError

    def exec(self, imgs, *args):
        if isinstance(imgs, abc.Sequence):
            ret = []
            if isinstance(imgs[0], abc.Sequence):
                basket = []
                for v in zip(*imgs):
                    basket.append(self.exec(v, *args))
                for r in zip(*basket):
                    ret.append(list(r))
            else:
                if self.pair_fn is None:
                    for i in imgs:
                        ret.append(self.call(i, *args))
                else:
                    params = self.pair_fn(args)
                    for idx, i in enumerate(imgs):
                        ret.append(self.call(i, *(params[idx])))
        else:
            ret = self.call(imgs, *args)

        return ret



class HWC2CHW(object):
    def __init__(self):
        super(HWC2CHW, self).__init__()

    def __call__(self, img):
        return np.transpose(img, (2, 0, 1))



class Random(object):
    def __init__(self, p, comburant):
        super(Random, self).__init__()
        if isinstance(p, (tuple, list)):
            assert len(p)==len(comburant), 'NEBULAE ERROR ⨷ the number of prob does not match with comburants.'
            assert sum(p)==1, 'NEBULAE ERROR ⨷ the sum of probabilities must be 1.'
        self.p = p
        self.cbr = comburant

    def __call__(self, imgs):
        if not isinstance(self.p, (tuple, list)):
            if rand.random() < self.p:
                self.cbr._format = self._format
                return self.cbr(imgs)
            else:
                return imgs
        else:
            cbr = np.random.choice(self.cbr, size=1, replace=False, p=self.p)[0]
            cbr._format = self._format
            return cbr(imgs)



def multiple(scales):
    if isinstance(scales, abc.Sequence):
        def _multiply(x, s):
            if isinstance(x, abc.Sequence):
                ret = []
                for v in x:
                    ret.append(_multiply(v, s))
                return ret
            else:
                return x * s
        return lambda x: [_multiply(x, s) for s in scales]
    else:
        raise TypeError('NEBULAE ERROR ⨷ non-iterable scales does not make sense.')



class Void():
    def __init__(self):
        pass

    def __call__(self, imgs):
        return imgs



class Pad(ABC):
    def __init__(self, padding, pair_fn=None):
        # padding: (top, bottom, left, right)
        super(Pad, self).__init__(pair_fn)
        self.padding = padding

    def _pad_np(self, img, padding):
        if img.ndim == 2:
            img = np.pad(img, ((padding[0], padding[1]), (padding[2], padding[3])), constant_values=0.)
        elif img.ndim == 3:
            img = np.pad(img, ((padding[0], padding[1]), (padding[2], padding[3]), (0, 0)), constant_values=0.)
        else:
            raise AssertionError('NEBULAE ERROR ⨷ the dimension of an image is either 2 or 3.')
        return img

    def _pad_pil(self, img, padding):
        padding = (padding[2], padding[0], padding[3], padding[1])
        return F.pad(img, padding, 0, 'constant')

    def call(self, img, padding):
        if self._format == NUMPY:
            return self._pad_np(img, padding)
        elif self._format == PIL:
            return self._pad_pil(img, padding)

    def __call__(self, imgs):
        ret = self.exec(imgs, self.padding)
        return ret



class Resize(ABC):
    def __init__(self, size, interp=LINEAR, pair_fn=None):
        # size: (height, width)
        super(Resize, self).__init__(pair_fn)
        self.size = size
        self.interp = interp

    def _resize_np(self, img, w, h):
        return cv2.resize(img, (w, h), interpolation=cv2.INTER_LINEAR)#CV_INTER[self.interp])

    def _resize_pil(self, img, w, h):
        return img.resize((w, h), PIL_INTERP[self.interp])

    def call(self, img, w, h):
        if self._format == NUMPY:
            return self._resize_np(img, w, h)
        elif self._format == PIL:
            return self._resize_pil(img, w, h)

    def __call__(self, imgs):
        ret = self.exec(imgs, self.size[1], self.size[0])
        return ret



class Crop(ABC):
    def __init__(self, size, area_ratio=(1, 1), aspect_ratio=(1, 1), central=False, interp=LINEAR, pair_fn=None):
        # size: (height, width)
        super(Crop, self).__init__(pair_fn)
        self.size = size
        self.area_ratio = area_ratio
        self.aspect_ratio = aspect_ratio
        self.central = central
        self.interp = interp
        if area_ratio == aspect_ratio == (1,1):
            self._reshape = False
        else: # firstly crop a patch that meets area and aspect ratio, then resize it to the given size
            self._reshape = True

    def _get_param_np(self, ref):
        height, width = ref.shape[:2]
        height = max(height, self.size[0])
        width = max(width, self.size[1])
        if self._reshape:
            sqrt_aspect_ratio = math.sqrt(rand.random() *
                                          (self.aspect_ratio[1] - self.aspect_ratio[0]) + self.aspect_ratio[0])
            sqrt_area_ratio = math.sqrt(rand.random() *
                                        (self.area_ratio[1] - self.area_ratio[0]) + self.area_ratio[0])
            for _ in range(10):
                w = int(sqrt_area_ratio * width * sqrt_aspect_ratio)
                h = int(sqrt_area_ratio * height / sqrt_aspect_ratio)
                if 0 < w <= width and 0 < h <= height:
                    if self.central:
                        x = (width - w) // 2
                        y = (height - h) // 2
                    else:
                        y = rand.randint(0, height - h)
                        x = rand.randint(0, width - w)
                    return y, x, h, w
            # Fallback to central crop
            in_ratio = float(width) / float(height)
            if (in_ratio < self.aspect_ratio[0]):
                w = width
                h = int(round(w / self.aspect_ratio[0]))
            elif (in_ratio > self.aspect_ratio[1]):
                h = height
                w = int(round(h * self.aspect_ratio[1]))
            else: # whole image
                w = width
                h = height
            y = (height - h) // 2
            x = (width - w) // 2
            return y, x, h, w
        else:
            w = self.size[1]
            h = self.size[0]
            if self.central:
                x = (width - w) // 2
                y = (height - h) // 2
            else:
                x = int(rand.random() * (width - w + 1))
                y = int(rand.random() * (height - h + 1))
            return y, x, h, w

    def _crop_np(self, img, y, x, h, w, size):
        height, width = img.shape[:2]
        if img.ndim == 2:
            # pad the width if needed
            if width < w:
                half = (w - width) // 2
                img = np.pad(img, ((0, 0), (half, w - width - half)), constant_values=0.)
            # pad the height if needed
            if height < h:
                half = (h - height) // 2
                img = np.pad(img, ((half, h - height - half), (0, 0)), constant_values=0.)
        elif img.ndim == 3:
            # pad the width if needed
            if width < w:
                half = (w - width) // 2
                img = np.pad(img, ((0, 0), (half, w - width - half), (0, 0)), constant_values=0.)
            # pad the height if needed
            if height < h:
                half = (h - height) // 2
                img = np.pad(img, ((half, h - height - half), (0, 0), (0, 0)), constant_values=0.)
        else:
            raise AssertionError('NEBULAE ERROR ⨷ the dimension of an image is either 2 or 3.')

        if self._reshape:
            cropped = cv2.resize(img[y:y + h, x:x + w], (size[1], size[0]), interpolation=CV_INTER[self.interp])
        else:
            cropped = img[y:y + h, x:x + w]
        return cropped

    def _get_param_pil(self, ref):
        # pad the width if needed
        if ref.size[0] < self.size[1]:
            ref = F.pad(ref, (self.size[1] - ref.size[0], 0), 0, 'constant')
        # pad the height if needed
        if ref.size[1] < self.size[0]:
            ref = F.pad(ref, (0, self.size[0] - ref.size[1]), 0, 'constant')
        if self._reshape:
            self.comburant = RandomResizedCrop(self.size, self.area_ratio, self.aspect_ratio, PIL_INTERP[self.interp])
            param = self.comburant.get_params(ref, self.comburant.scale, self.comburant.ratio)
        else:
            self.comburant = RandomCrop(self.size)
            param = self.comburant.get_params(ref, self.comburant.size)
        return param

    def _crop_pil(self, img, y, x, h, w, size):
        # pad the width if needed
        if img.size[0] < size[1]:
            img = F.pad(img, (size[1] - img.size[0], 0), 0, 'constant')
        # pad the height if needed
        if img.size[1] < size[0]:
            img = F.pad(img, (0, size[0] - img.size[1]), 0, 'constant')

        if self._reshape:
            return F.resized_crop(img, y, x, h, w, size, self.comburant.interpolation)
        else:
            if self.central:
                return F.center_crop(img, size)
            else:
                return F.crop(img, y, x, h, w)

    def call(self, img, y, x, h, w, size):
        if self._format == NUMPY:
            return self._crop_np(img, y, x, h, w, size)
        elif self._format == PIL:
            return self._crop_pil(img, y, x, h, w, size)

    def __call__(self, imgs):
        if self._format == NUMPY:
            img = imgs
            while isinstance(img, abc.Sequence):
                img = img[0]
            y, x, h, w = self._get_param_np(img)
        else:
            img = imgs
            while isinstance(img, abc.Sequence):
                img = img[0]
            y, x, h, w = self._get_param_pil(img) # it has no effect if central is True

        ret = self.exec(imgs, y, x, h, w, self.size)
        return ret



class Flip(ABC):
    def __init__(self, axial):
        super(Flip, self).__init__()
        self.axial = axial

    def _flip_np(self, img):
        if self.axial == HORIZONTAL:
            dst = np.flip(img, 0)
        elif self.axial == VERTICAL:
            dst = np.flip(img, 1)
        else:
            raise Exception('NEBULAE ERROR ⨷ the invoked flip type is not defined or supported.')
        return dst

    def _flip_pil(self, img):
        if self.axial == HORIZONTAL:
            comburant = RandomVerticalFlip(1)
        elif self.axial == VERTICAL:
            comburant = RandomHorizontalFlip(1)
        else:
            raise Exception('NEBULAE ERROR ⨷ the invoked flip type is not defined or supported.')
        return comburant(img)

    def call(self, img):
        if self._format == NUMPY:
            return self._flip_np(img)
        elif self._format == PIL:
            return self._flip_pil(img)

    def __call__(self, imgs):
        ret = self.exec(imgs)
        return ret



class Rotate(ABC):
    def __init__(self, degree, intact=False, interp=NEAREST):
        '''
        Args
        intact: whether to keep image intact which might enlarge the output size
        '''
        super(Rotate, self).__init__()
        if isinstance(degree, (tuple, list)):
            assert degree[1] >= degree[0], 'NEBULAE ERROR ⨷ the second element should not be less than the first.'
        elif isinstance(degree, (int, float)):
            assert abs(degree) <= 360, 'NEBULAE ERROR ⨷ a valid degree should have its absolute value under 360.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid factor should be an float or a tuple.')
        self.degree = degree
        self.intact = intact
        self.interp = interp

    def _rotate_np(self, img, degree):
        h, w = img.shape[:2]
        mtx = cv2.getRotationMatrix2D((w / 2, h / 2), degree, 1)
        if self.intact:
            rot_vtx = []
            vertices = [np.array([0., 0.]), np.array([0., h]), np.array([w, 0.]), np.array([float(w), float(h)])]
            for v in vertices:
                rot_vtx.append(mtx[:, :2] @ v.T + mtx[:, 2])
            rot_vtx = np.stack(rot_vtx, axis=1)
            w_prime = math.ceil(rot_vtx[0].max() - rot_vtx[0].min())
            h_prime = math.ceil(rot_vtx[1].max() - rot_vtx[1].min())
            mtx[0, -1] += (w_prime - w) / 2
            mtx[1, -1] += (h_prime - h) / 2
            dst = cv2.warpAffine(img, mtx, (w_prime, h_prime))
        else:
            dst = cv2.warpAffine(img, mtx, (w, h))
        return dst

    def _rotate_pil(self, img, degree):
        return F.rotate(img, degree, self.comburant.resample, self.comburant.expand,
                        self.comburant.center, self.comburant.fill)

    def call(self, img, degree):
        if self._format == NUMPY:
            return self._rotate_np(img, degree)
        elif self._format == PIL:
            return self._rotate_pil(img, degree)

    def __call__(self, imgs):
        if self._format == NUMPY:
            if isinstance(self.degree, (tuple, list)):
                degree = rand.random() * (self.degree[1] - self.degree[0]) + self.degree[0]
            else:
                degree = self.degree
        else:
            self.comburant = RandomRotation(self.degree, PIL_INTERP[self.interp], self.intact)
            if isinstance(self.degree, (tuple, list)):
                degree = self.comburant.get_params(self.comburant.degrees)
            else:
                degree = self.degree
        ret = self.exec(imgs, degree)
        return ret



class Brighten(ABC):
    def __init__(self, factor):
        super(Brighten, self).__init__()
        if isinstance(factor, (tuple, list)):
            assert factor[1] >= factor[0] and factor[0] >= 0, \
                'NEBULAE ERROR ⨷ the second element should not be less than the first one and they are both non-negative.'
        elif isinstance(factor, (int, float)):
            assert factor >= 0, 'NEBULAE ERROR ⨷ a valid factor should be non-negative.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid factor should be an float or a tuple.')
        self.factor = factor

    def _brighten_np(self, img, factor):
        dst = np.round(img * float(factor)) # pixels range between [theta, 1+theta]
        dst = np.clip(dst, 0, 255).astype(np.uint8) # pixels range between [0, 1]
        return dst

    def _brighten_pil(self, img, factor):
        return F.adjust_brightness(img, factor)

    def call(self, img, factor):
        if self._format == NUMPY:
            return self._brighten_np(img, factor)
        elif self._format == PIL:
            return self._brighten_pil(img, factor)

    def __call__(self, imgs):
        if isinstance(self.factor, (tuple, list)):
            factor = rand.uniform(self.factor[0], self.factor[1])
        else:
            factor = self.factor
        ret = self.exec(imgs, factor)
        return ret



class Contrast(ABC):
    def __init__(self, factor):
        super(Contrast, self).__init__()
        if isinstance(factor, (tuple, list)):
            assert factor[1] >= factor[0] and factor[0] >= 0, \
                'NEBULAE ERROR ⨷ the second element should not be less than the first one and they are both non-negative.'
        elif isinstance(factor, (int, float)):
            assert factor >= 0, 'NEBULAE ERROR ⨷ a valid factor should be non-negative.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid factor should be an float or a tuple.')
        self.factor = factor

    def _contrast_np(self, img, factor):
        if img.ndim == 3:
            gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        else:
            gray = img
        white = np.ones_like(img)
        dst = cv2.addWeighted(img, factor, white, gray.mean() * (1. - factor), 0)
        return dst

    def _contrast_pil(self, img, factor):
        return F.adjust_contrast(img, factor)

    def call(self, img, factor):
        if self._format == NUMPY:
            return self._contrast_np(img, factor)
        elif self._format == PIL:
            return self._contrast_pil(img, factor)

    def __call__(self, imgs):
        if isinstance(self.factor, (tuple, list)):
            factor = rand.uniform(self.factor[0], self.factor[1])
        else:
            factor = self.factor
        ret = self.exec(imgs, factor)
        return ret



class Saturate(ABC):
    def __init__(self, factor):
        super(Saturate, self).__init__()
        if isinstance(factor, (tuple, list)):
            assert factor[1] >= factor[0] and factor[0] >= 0, \
                'NEBULAE ERROR ⨷ the second element should not be less than the first one and they are both non-negative.'
        elif isinstance(factor, (int, float)):
            assert factor >= 0, 'NEBULAE ERROR ⨷ a valid factor should be non-negative.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid factor should be an float or a tuple.')
        self.factor = factor

    def _saturate_np(self, img, factor):
        if img.ndim == 3:
            gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        else:
            gray = img
        gray = cv2.cvtColor(gray, cv2.COLOR_GRAY2RGB)
        dst = cv2.addWeighted(img, factor, gray, 1. - factor, 0)
        return dst

    def _saturate_pil(self, img, factor):
        return F.adjust_saturation(img, factor)

    def call(self, img, factor):
        if self._format == NUMPY:
            return self._saturate_np(img, factor)
        elif self._format == PIL:
            return self._saturate_pil(img, factor)

    def __call__(self, imgs):
        if isinstance(self.factor, (tuple, list)):
            factor = rand.uniform(self.factor[0], self.factor[1])
        else:
            factor = self.factor
        ret = self.exec(imgs, factor)
        return ret



class Hue(ABC):
    def __init__(self, factor):
        # hue range: [-0.5, 0.5]
        super(Hue, self).__init__()
        if isinstance(factor, (tuple, list)):
            assert factor[1] >= factor[0] and factor[0] >= -0.5 and factor[1] <= 0.5, \
                'NEBULAE ERROR ⨷ the second element should not be less than the first one' \
                + ' and their absolute values are less than 0.5.'
        elif isinstance(factor, (int, float)):
            assert abs(factor) <= 0.5, 'NEBULAE ERROR ⨷ a valid degree should have its absolute value under 0.5.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid factor should be an float or a tuple.')
        self.factor = factor

    def _hue_np(self, img, factor):
        if img.ndim == 3:
            hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV_FULL)
            hsv[:, :, 0] += (factor * 255 * np.ones(hsv.shape[:2])).astype(np.uint8)
            return cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB_FULL)
        else:
            return img

    def _hue_pil(self, img, factor):
        return F.adjust_hue(img, factor)

    def call(self, img, factor):
        if self._format == NUMPY:
            return self._hue_np(img, factor)
        elif self._format == PIL:
            return self._hue_pil(img, factor)

    def __call__(self, imgs):
        if isinstance(self.factor, (tuple, list)):
            factor = rand.uniform(self.factor[0], self.factor[1])
        else:
            factor = self.factor
        ret = self.exec(imgs, factor)
        return ret



class Blur(ABC):
    def __init__(self, radius):
        super(Blur, self).__init__()
        if isinstance(radius, tuple):
            assert radius[1] >= radius[0], 'NEBULAE ERROR ⨷ the second element should not be less than the first.'
        elif isinstance(radius, int):
            assert radius>=1, 'NEBULAE ERROR ⨷ a valid radius should be larger than one.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid radius should be an integer or a tuple.')
        self.radius = radius

    def _blur_np(self, img, radius):
        diameter = 2 * radius + 1
        return cv2.GaussianBlur(img, (diameter, diameter), 0)

    def _blur_pil(self, img, radius):
        return img.filter(ImageFilter.GaussianBlur(radius))

    def call(self, img, factor):
        if self._format == NUMPY:
            return self._blur_np(img, factor)
        elif self._format == PIL:
            return self._blur_pil(img, factor)

    def __call__(self, imgs):
        if isinstance(self.radius, tuple):
            radius = rand.randint(self.radius[0], self.radius[1])
        else:
            radius = self.radius
        ret = self.exec(imgs, radius)
        return ret


class Sharpen(ABC):
    def __init__(self, factor):
        super(Sharpen, self).__init__()
        if isinstance(factor, tuple):
            assert factor[1] >= factor[0], 'NEBULAE ERROR ⨷ the second element should not be less than the first.'
        elif isinstance(factor, (int, float)):
            assert factor>1, 'NEBULAE ERROR ⨷ a valid factor should be larger than one.'
        else:
            raise TypeError('NEBULAE ERROR ⨷ a valid factor should be an float or a tuple.')
        self.factor = factor

    def _sharpen_np(self, img, factor):
        radius = min(3, int(factor))
        diameter = 2 * radius + 1
        blurred = cv2.GaussianBlur(img, (diameter, diameter), 0)
        return cv2.addWeighted(img, factor, blurred, 1-factor, 0)

    def _sharpen_pil(self, img, factor):
        enhancer = ImageEnhance.Sharpness(img)
        return enhancer.enhance(factor)

    def call(self, img, factor):
        if self._format == NUMPY:
            return self._sharpen_np(img, factor)
        elif self._format == PIL:
            return self._sharpen_pil(img, factor)

    def __call__(self, imgs):
        if isinstance(self.factor, tuple):
            factor = self.factor[0] + rand.random()*(self.factor[1]-self.factor[0])
        else:
            factor = self.factor
        ret = self.exec(imgs, factor)
        return ret



class Noise(ABC):
    def __init__(self, theta, distrib):
        super(Noise, self).__init__()
        self.theta = theta
        assert distrib in (GAUSSIAN, POISSON), 'NEBULAE ERROR ⨷ the distribution is either to be Gaussian or Poisson.'
        self.distrib = distrib

    def _noise_np(self, img, theta):
        h, w, c = img.shape
        if self.distrib == GAUSSIAN:
            noise = (np.random.normal(scale=self.theta, size=(h, w, c)) * 255).astype(np.uint8)
        else:
            noise = np.random.poisson(img * theta).astype(np.uint8)
        img = cv2.addWeighted(img, 1, noise, 1, 0)
        return img

    def _noise_pil(self, img, theta):
        img = np.array(img)
        img = self._noise_np(img, theta)
        img = Image.fromarray(img)
        return img

    def call(self, img, theta):
        if self._format == NUMPY:
            return self._noise_np(img, theta)
        elif self._format == PIL:
            return self._noise_pil(img, theta)

    def __call__(self, imgs):
        if isinstance(self.theta, tuple):
            theta = self.theta[0] + rand.random()*(self.theta[1]-self.theta[0])
        else:
            theta = self.theta
        ret = self.exec(imgs, theta)
        return ret



class Sinc(ABC):
    def __init__(self, omega, diameter, ret_img=True):
        # >| omega: usually between (0, pi]
        super(Sinc, self).__init__()
        self.omega = omega
        self.diameter = diameter
        self.ret_img = ret_img

    def _sinc_np(self, img, omega, diameter):
        cx = cy = (diameter - 1) / 2
        with np.errstate(divide='ignore', invalid='ignore'):
            kernel = np.fromfunction(lambda x,y :
                                     omega*special.j1(omega*np.sqrt((x-cx)**2 + (y-cy)**2))
                                     / (2*np.pi*np.sqrt((x-cx)**2 + (y-cy)**2)),
                                        [diameter, diameter])
        if diameter % 2:
            kernel[(diameter-1)//2, (diameter-1)//2] = omega**2 / (4*np.pi)
        kernel /= np.sum(kernel)
        if self.ret_img:
            img = img.astype(np.float32)
            if img.ndim == 3:
                arr = np.zeros_like(img)
                for c in range(img.shape[-1]):
                    arr[:, :, c] = signal.convolve2d(img[:, :, c], kernel, mode='same', boundary='symm')
                img = np.clip(arr, 0, 255)
            else:
                img = np.clip(signal.convolve2d(img, kernel, mode='same', boundary='symm'), 0, 255)
            return img.astype(np.uint8)
        else:
            if img.ndim == 3:
                return np.tile(np.expand_dims(kernel, -1), (1, 1, 3))
            else:
                return kernel

    def _sinc_pil(self, img, omega, diameter):
        img = np.array(img)
        img = self._sinc_np(img, omega, diameter)
        if self.ret_img:
            dst = Image.fromarray(img)
        else:
            dst = img
        return dst

    def call(self, img, omega, diameter):
        if self._format == NUMPY:
            return self._sinc_np(img, omega, diameter)
        elif self._format == PIL:
            return self._sinc_pil(img, omega, diameter)

    def __call__(self, imgs):
        if isinstance(self.omega, tuple):
            omega = rand.uniform(self.omega[0], self.omega[1])
        else:
            omega = self.omega
        if isinstance(self.diameter, tuple):
            diameter = rand.randint(self.diameter[0], self.diameter[1])
        else:
            diameter = self.diameter
        ret = self.exec(imgs, omega, diameter)
        return ret



class JPEG(ABC):
    def __init__(self, quality):
        super(JPEG, self).__init__()
        if isinstance(quality, tuple):
            assert quality[1] >= quality[0], 'NEBULAE ERROR ⨷ the second element should not be less than the first.'
            assert quality[0]>0 and quality[1]<100, 'NEBULAE ERROR ⨷ a valid quality should be an integer within [1, 99].'
        elif isinstance(quality, int):
            assert quality>0 and quality<100, 'NEBULAE ERROR ⨷ a valid quality should be an integer within [1, 99].'
        else:
            raise ValueError('NEBULAE ERROR ⨷ a valid quality should be an integer or a tuple.')
        self.quality = quality

    def _jpeg_np(self, img, qlt):
        _, img = cv2.imencode('.jpg', img, [cv2.IMWRITE_JPEG_QUALITY, qlt])
        dst = cv2.imdecode(img, cv2.IMREAD_UNCHANGED)
        return dst

    def _jpeg_pil(self, img, qlt):
        buffer = BytesIO()
        img.save(buffer, 'JPEG', quality=qlt)
        return Image.open(buffer)

    def call(self, img, qlt):
        if self._format == NUMPY:
            return self._jpeg_np(img, qlt)
        elif self._format == PIL:
            return self._jpeg_pil(img, qlt)

    def __call__(self, imgs):
        if isinstance(self.quality, tuple):
            qlt = int(rand.randint(self.quality[0], self.quality[1]) * 0.96 + 0.1)
        else:
            qlt = int(self.quality * 0.96 + 0.1)
        ret = self.exec(imgs, qlt)
        return ret



class WebP(ABC):
    def __init__(self, quality):
        super(WebP, self).__init__()
        if isinstance(quality, tuple):
            assert quality[1] >= quality[0], 'NEBULAE ERROR ⨷ the second element should not be less than the first.'
            assert quality[0]>0 and quality[1]<100, 'NEBULAE ERROR ⨷ a valid quality should be an integer within [1, 99].'
        elif isinstance(quality, int):
            assert quality>0 and quality<100, 'NEBULAE ERROR ⨷ a valid quality should be an integer within [1, 99].'
        else:
            raise ValueError('NEBULAE ERROR ⨷ a valid quality should be an integer or a tuple.')
        self.quality = quality

    def _webp_np(self, img, qlt):
        _, img = cv2.imencode('.webp', img, [cv2.IMWRITE_WEBP_QUALITY, qlt])
        dst = cv2.imdecode(img, cv2.IMREAD_UNCHANGED)
        return dst

    def _webp_pil(self, img, qlt):
        buffer = BytesIO()
        img.save(buffer, 'WebP', quality=qlt)
        return Image.open(buffer)

    def call(self, img, qlt):
        if self._format == NUMPY:
            return self._png_np(img, qlt)
        elif self._format == PIL:
            return self._png_pil(img, qlt)

    def __call__(self, imgs):
        if isinstance(self.quality, tuple):
            qlt = rand.randint(self.quality[0], self.quality[1])
        else:
            qlt = self.quality
        ret = self.exec(imgs, qlt)
        return ret



def _vid_encode(imgs, fps, bitrate, codec, _format):
    buf = BytesIO()
    with av.open(buf, 'w', 'mp4') as container:
        stream = container.add_stream(codec, rate=fps)
        if _format == NUMPY:
            h, w = imgs[0].shape[:2]
        elif _format == PIL:
            w, h = imgs[0].size
        else:
            raise Exception('NEBULAE ERROR ⨷ the image format has not been assigned.')
        stream.height = h
        stream.width = w
        stream.pix_fmt = 'yuv420p'
        stream.bit_rate = bitrate

        for img in imgs:
            if _format == NUMPY:
                frame = av.VideoFrame.from_ndarray(img, format='rgb24')
            elif _format == PIL:
                frame = av.VideoFrame.from_image(img)
            frame.pict_type = 'NONE'
            for packet in stream.encode(frame):
                container.mux(packet)

        # Flush stream
        for packet in stream.encode():
            container.mux(packet)

    outputs = []
    with av.open(buf, 'r', 'mp4') as container:
        if container.streams.video:
            if _format == NUMPY:
                for frame in container.decode(**{'video': 0}):
                    outputs.append(frame.to_rgb().to_ndarray())
            else:
                for frame in container.decode(**{'video': 0}):
                    outputs.append(frame.to_rgb().to_image())
    return outputs



class VABC(ABC):
    def __init__(self, fps, br):
        super(VABC, self).__init__()
        self.fps = fps
        self.br = br

    def __call__(self, seqs):
        if isinstance(self.fps, tuple):
            fps = rand.randint(self.fps[0], self.fps[1])
        else:
            fps = self.fps
        if isinstance(self.br, tuple):
            br = rand.randint(self.br[0], self.br[1])
        else:
            br = self.br

        if isinstance(seqs[0], abc.Sequence):
            ret = []
            for s in seqs:
                ret.append(self.call(s, fps, br))
        else:
            ret = self.call(seqs, fps, br)

        return ret



class MPEG4(VABC):
    def __init__(self, fps, br):
        super(MPEG4, self).__init__(fps, br)

    def call(self, imgs, fps, br):
        return _vid_encode(imgs, fps, br, 'mpeg4', self._format)



class H264(VABC):
    def __init__(self, fps, br):
        super(H264, self).__init__(fps, br)

    def call(self, imgs, fps, br):
        return _vid_encode(imgs, fps, br, 'h264', self._format)



class VP9(VABC):
    def __init__(self, fps, br):
        super(VP9, self).__init__(fps, br)

    def call(self, imgs, fps, br):
        return _vid_encode(imgs, fps, br, 'vp9', self._format)



class AV1(VABC):
    def __init__(self, fps, br):
        super(AV1, self).__init__(fps, br)

    def call(self, imgs, fps, br):
        return _vid_encode(imgs, fps, br, 'av1', self._format)



if __name__ == '__main__':
    from time import time
    import cv2
    from PIL import ImageEnhance
    # >| VABC test
    seq = []
    for i in range(5):
        img = cv2.imread('/Users/Seria/Desktop/nebulae/test/vid/%04d.png'%(i+1))
        img = img[:,:,::-1]
        seq.append(img)
    cbr = Comburant(H264(30, 6e5))
    dst = cbr(seq)
    for i in range(5):
        # dst[i].save('/Users/Seria/Desktop/nebulae/test/img/%04d.png'%(i+1), format='PNG', compress_level=0)
        cv2.imwrite('/Users/Seria/Desktop/nebulae/test/img/np-lr-%03d.png' % (i + 1),
                    dst[i][:, :, ::-1], [cv2.IMWRITE_PNG_COMPRESSION, 0])


    # >| ABC test
    cbr = Comburant(Crop((128, 128), central=True, pair_fn=multiple((1, 2))),
                    Resize((100, 100), pair_fn=multiple((1, 2))),
                    Flip(HORIZONTAL),
                    Rotate(30, True),
                    Brighten(1.2),
                    Contrast(1.5),
                    Saturate(0.6),
                    Hue(0.3),
                    Blur(1),
                    Sharpen(1.5),
                    Noise(0.2, POISSON),
                    Sinc(np.pi/3, 15),
                    JPEG(90),
                    )#format=PIL, end=False)
    t_ = time()
    niter = 20
    lrs = []
    hrs = []
    for i in range(niter):
        path = '/Users/Seria/Desktop/nebulae/test/vid/%04d.png'%(i+1)

        # img = Image.open(path)
        # hrs.append(img)
        # lrs.append(img.resize((img.size[0] // 2, img.size[1] // 2)))

        img = cv2.imread(path)
        img = img[:,:,::-1]
        hrs.append(img)
        lrs.append(cv2.resize(img, (img.shape[1]//2, img.shape[0]//2)))

    imgs = cbr([lrs, hrs])
    _t = time()
    for i in range(1):
        # imgs[0][i].save('/Users/Seria/Desktop/nebulae/test/img/pil-lr-%03d.png' % (i + 1), format='PNG', compress_level=0)
        # imgs[1][i].save('/Users/Seria/Desktop/nebulae/test/img/pil-hr-%03d.png' % (i + 1), format='PNG', compress_level=0)
        cv2.imwrite('/Users/Seria/Desktop/nebulae/test/img/np-lr-%03d.png'%(i+1), imgs[0][i][:,:,::-1])#, [cv2.IMWRITE_PNG_COMPRESSION, 0])
        cv2.imwrite('/Users/Seria/Desktop/nebulae/test/img/np-hr-%03d.png'%(i+1), imgs[1][i][:,:,::-1])#, [cv2.IMWRITE_PNG_COMPRESSION, 0])
    print((_t-t_)/niter)


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:57 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import cv2
import os
import time

def Timer(func):
    def _wrapper(*args, **kwargs):
        t_ = time.time()
        ret = func(*args, **kwargs)
        _t = time.time()
        if not isinstance(ret, tuple):
            ret = (ret,)
        return (_t - t_,) + ret

    return _wrapper


def SPST(func):
    def _wrapper(*args, **kwargs):
        cv2.setNumThreads(0)
        cv2.ocl.setUseOpenCL(False)
        return func(*args, **kwargs)

    return _wrapper


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 2:50 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from .decorator import Timer, SPST
from .utility import hotvec2mtx, mtx2hotvec, den2spa, spa2den, rand_trunc_norm, parse_cfg, record_cfg, \
                    byte2arr, rgb2y, join_imgs, plot_in_one, ver2num, GPUtil, sprawl


import torch
import random
import numpy as np
def destine(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)


#!/usr/bin/env python
'''
architect
Created by Seria at 2022/11/7 12:08 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock, Craft
import numpy as np
from math import ceil


BN = 20
CBN = 21
IN = 22
CIN = 23
LN = 24



class MLPE(Craft):
    def __init__(self, in_shape, hidden_dim, latent_dim, scope='MLPE'):
        super(MLPE, self).__init__(scope)
        H, W, C = in_shape
        in_dim = H * W * C
        self.in_dim = in_dim
        self.rsp = dock.Reshape()
        self.hidden = dock.Dense(in_dim, hidden_dim)
        self.relu = dock.Relu()
        self.mu = dock.Dense(hidden_dim, latent_dim)
        self.log_var = dock.Dense(hidden_dim, latent_dim)
        self.exp = dock.Exp()

    def reparam(self, mu, lv):
        eps = dock.coat(np.random.normal(size=mu.shape).astype(np.float32))
        return mu + self.exp(lv / 2) * eps

    def run(self, x):
        x = self.rsp(x, (-1, self.in_dim))
        x = self.hidden(x)
        x = self.relu(x)
        mu = self.mu(x)
        lv = self.log_var(x)
        z = self.reparam(mu, lv)
        return mu, lv, z


class MLPD(Craft):
    def __init__(self, out_shape, hidden_dim, latent_dim, scope='MLPD'):
        super(MLPD, self).__init__(scope)
        H, W, C = out_shape
        out_dim = H * W * C
        self.out_shape = tuple(out_shape[-1:] + out_shape[:2])
        self.hidden = dock.Dense(latent_dim, hidden_dim)
        self.relu = dock.Relu()
        self.fc = dock.Dense(hidden_dim, out_dim)
        self.tanh = dock.Tanh()
        self.rsp = dock.Reshape()

    def run(self, z):
        z = self.hidden(z)
        z = self.relu(z)
        z = self.fc(z)
        z = self.tanh(z)
        x = self.rsp(z, (-1,) + self.out_shape)
        return x




class DeconvD(Craft):
    def __init__(self, out_shape, latent_dim, scope='DECONVD'):
        super(DeconvD, self).__init__(scope)
        H, W, C = out_shape
        assert H%4==0 and W%4==0
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()

        self.fc_1 = dock.Dense(latent_dim, 128 * (H // 4) * (W // 4))
        self.bn_1 = dock.BN(128, dim=2)
        pad = dock.autopad((3, 3), stride=2, size=(H // 2, W // 2))
        self.conv_1 = dock.TransConv(128, 128, (H // 2, W // 2), kernel=(3, 3), stride=2, padding=pad)
        self.bn_2 = dock.BN(128, dim=2, mmnt=0.8)
        pad = dock.autopad((3, 3), stride=2, size=(H, W))
        self.conv_2 = dock.TransConv(128, 64, (H, W), kernel=(3, 3), stride=2, padding=pad)
        self.bn_3 = dock.BN(64, dim=2, mmnt=0.8)
        pad = dock.autopad((3, 3), size=(H, W))
        self.conv_3 = dock.Conv(64, C, (3, 3), padding=pad)

        self.tanh = dock.Tanh()
        self.reshape = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc_1(self['latent_code'])
        z = self.reshape(z, (-1, 128, self.H//4, self.W//4))

        z = self.bn_1(z)
        z = self.conv_1(z)

        z = self.bn_2(z)
        z = self.lrelu(z)
        z = self.conv_2(z)

        z = self.bn_3(z)
        z = self.lrelu(z)
        z = self.conv_3(z)

        self['fake'] = self.tanh(z)

        return self['fake']


class ConvE(Craft):
    def __init__(self, in_shape, scope='CONVE'):
        super(ConvE, self).__init__(scope)
        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()
        self.flat = dock.Reshape()
        self.exp = dock.Exp()
        pad = dock.autopad((3, 3), stride=2, size=(H, W))
        self.conv_1 = dock.Conv(C, 16, (3, 3), stride=2, padding=pad)
        pad = dock.autopad((3, 3), stride=2, size=(ceil(H / 2), ceil(W / 2)))
        self.conv_2 = dock.Conv(16, 32, (3, 3), stride=2, padding=pad)
        self.bn_2 = dock.BN(32, dim=2, mmnt=0.8)
        pad = dock.autopad((3, 3), stride=2, size=(ceil(H / 4), ceil(W / 4)))
        self.conv_3 = dock.Conv(32, 64, (3, 3), stride=2, padding=pad)
        self.bn_3 = dock.BN(64, dim=2, mmnt=0.8)
        pad = dock.autopad((3, 3), stride=2, size=(ceil(H / 8), ceil(W / 8)))
        self.conv_m = dock.Conv(64, 128, (3, 3), stride=2, padding=pad)
        self.bn_m = dock.BN(128, dim=2, mmnt=0.8)
        pad = dock.autopad((3, 3), stride=2, size=(ceil(H / 8), ceil(W / 8)))
        self.conv_v = dock.Conv(64, 128, (3, 3), stride=2, padding=pad)
        self.bn_v = dock.BN(128, dim=2, mmnt=0.8)

    def reparam(self, mu, lv):
        eps = dock.coat(np.random.normal(size=mu.shape).astype(np.float32))
        return mu + self.exp(lv / 2) * eps

    def run(self, x):
        bs = x.shape[0]
        self['input'] = x
        x = self.conv_1(x)
        x = self.lrelu(x)

        x = self.conv_2(x)
        x = self.lrelu(x)
        x = self.bn_2(x)

        x = self.conv_3(x)
        x = self.lrelu(x)
        x = self.bn_3(x)

        mu = self.conv_m(x)
        mu = self.lrelu(mu)
        mu = self.bn_m(mu)
        mu = self.flat(mu, (bs, -1))

        lv = self.conv_v(x)
        lv = self.lrelu(lv)
        lv = self.bn_v(lv)
        lv = self.flat(lv, (bs, -1))

        z = self.reparam(mu, lv)

        return mu, lv, z









class ResBlock(Craft):
    def __init__(self, in_shape, neck_chs, body_chs, stride=1, norm_fn=BN,
                 attention=False, spec_norm=False, w_init=dock.XavierNorm(), scope='RESBLK'):
        super(ResBlock, self).__init__(scope)
        H, W, C = in_shape
        self.norm_fn = norm_fn
        self.body_chs = body_chs
        self.cn = False
        if norm_fn == BN:
            self.n_1 = dock.BN(C, dim=2)
            self.n_2 = dock.BN(neck_chs, dim=2)
            # self.ds_n = dock.BN(self.body_chs, dim=2)
        elif norm_fn == IN:
            self.n_1 = dock.IN(C, dim=2)
            self.n_2 = dock.IN(neck_chs, dim=2)
            # self.ds_n = dock.IN(self.body_chs, dim=2)
        elif norm_fn == CBN:
            self.n_1 = dock.CBN(128, C, dim=2)
            self.n_2 = dock.CBN(128, neck_chs, dim=2)
            # self.ds_n = dock.CBN(128, self.body_chs, dim=2)
            self.cn = True
        elif norm_fn == CIN:
            self.n_1 = dock.CIN(128, C, dim=2)
            self.n_2 = dock.CIN(128, neck_chs, dim=2)
            # self.ds_n = dock.CIN(128, self.body_chs, dim=2)
            self.cn = True
        elif norm_fn == LN:
            self.n_1 = dock.LN(in_shape)
            self.n_2 = dock.LN((ceil(H / stride), ceil(W / stride), neck_chs))
            # self.ds_n = dock.LN((ceil(H / stride), ceil(W / stride), self.body_chs))
        elif norm_fn is None:
            self.n_1 = dock.Identity()
            self.n_2 = dock.Identity()
            # self.ds_n = dock.Identity()

        pad = dock.autopad((3, 3), stride=stride, size=(H, W))
        if spec_norm:
            self.conv_1 = dock.SN(dock.Conv(C, neck_chs, (3, 3), stride=stride, padding=pad, w_init=w_init))
            self.conv_2 = dock.SN(dock.Conv(neck_chs, self.body_chs, (1, 1), w_init=w_init))
            self.ds_conv = dock.SN(dock.Conv(C, self.body_chs, (1, 1), stride, w_init=w_init))
        else:
            self.conv_1 = dock.Conv(C, neck_chs, (3, 3), stride=stride, padding=pad, w_init=w_init)
            self.conv_2 = dock.Conv(neck_chs, self.body_chs, (1, 1), w_init=w_init)
            self.ds_conv = dock.Conv(C, self.body_chs, (1, 1), stride, w_init=w_init)

        self.relu = dock.Relu()
        # if attention and H==W==64:
        #     self.att = NonLocal((H // stride, W // stride, self.body_chs), self.body_chs//8, self.body_chs//2,
        #                         spec_norm, w_init)
        self.att = dock.Identity()

    def run(self, x):
        if self.cn:
            self['input'], self['noise'] = x
            y = self.n_1(self['input'], self['noise'])
            y = self.relu(y)
            y = self.conv_1(y)

            y = self.n_2(y, self['noise'])
            y = self.relu(y)
            y = self.conv_2(y)

            self['identity'] = self.ds_conv(self['input'])
        else:
            self['input'] = x
            y = self.n_1(self['input'])
            y = self.relu(y)
            y = self.conv_1(y)

            y = self.n_2(y)
            y = self.relu(y)
            y = self.conv_2(y)

            self['identity'] = self.ds_conv(self['input'])

        y += self['identity']
        y = self.att(y)

        return y



class ResD(Craft):
    def __init__(self, in_shape, base_chs, latent_dim, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), scope='RESD'):
        super(ResD, self).__init__(scope)
        self.min_res = 4
        ichs = {64: [base_chs * c for c in (8, 8, 4, 2)],
                128: [base_chs * c for c in (16, 16, 8, 4, 2)],
                256: [base_chs * c for c in (16, 16, 8, 8, 4, 2)]}
        ochs = {64: [base_chs * c for c in (8, 4, 2, 1)],
                128: [base_chs * c for c in (16, 8, 4, 2, 1)],
                256: [base_chs * c for c in (16, 8, 8, 4, 2, 1)]}

        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C
        min_size = min(H, W)
        self.ichs = ichs[min_size]
        self.ochs = ochs[min_size]

        self.up = dock.Zoom(scale=(2, 2))
        self.relu = dock.Relu()
        pad = dock.autopad((3, 3), size=(H, W))
        if spec_norm:
            self.conv = dock.SN(dock.Conv(self.ochs[-1], C, (3, 3), padding=pad, w_init=w_init))
            self.fc = dock.SN(dock.Dense(latent_dim, self.ichs[0] * self.min_res * self.min_res, w_init=w_init))
        else:
            self.conv = dock.Conv(self.ochs[-1], C, (3, 3), padding=pad, w_init=w_init)
            self.fc = dock.Dense(latent_dim, self.ichs[0] * self.min_res * self.min_res, w_init=w_init)

        for i in range(len(self.ochs)):
            setattr(self, 'blk_%d'%i, ResBlock((self.min_res * 2**i, self.min_res * 2**i, self.ichs[i]),
                                                self.ichs[i]//2, self.ochs[i], norm_fn=norm_fn,
                                                attention=attention, spec_norm=spec_norm, w_init=w_init))

        self.cn = False
        if norm_fn == BN:
            self.nf = dock.BN(self.ochs[-1], dim=2)
        elif norm_fn == IN:
            self.nf = dock.IN(self.ochs[-1], dim=2)
        elif norm_fn == CBN:
            self.nf = dock.CBN(latent_dim, self.ochs[-1], dim=2)
            self.cn = True
        elif norm_fn == CIN:
            self.nf = dock.CIN(latent_dim, self.ochs[-1], dim=2)
            self.cn = True
        elif norm_fn == LN:
            self.nf = dock.LN((H, W, self.ochs[-1]))
        elif norm_fn is None:
            self.nf = dock.Identity()

        self.tanh = dock.Tanh()
        self.reshape = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc(self['latent_code'])
        z = self.reshape(z, (-1, self.ichs[0], self.min_res, self.min_res))

        if self.cn:
            for i in range(len(self.ochs)):
                rb = getattr(self, 'blk_%d' % i)
                z = rb((z, self['latent_code']))
                n_shape = z.shape
                n_shape[1] = 1
                noise = dock.coat(np.random.normal(size=n_shape).astype('float32'))
                strength = dock.coat(np.array(0).astype('float32'), as_const=False)
                z += strength * noise
                z = self.up(z)

            z = self.nf(z, self['latent_code'])
        else:
            for i in range(len(self.ochs)):
                rb = getattr(self, 'blk_%d' % i)
                z = rb(z)
                z = self.up(z)

            z = self.nf(z)

        z = self.relu(z)
        z = self.conv(z)

        self['fake'] = self.tanh(z)

        return self['fake']

class ResE(Craft):
    def __init__(self, in_shape, base_chs, latent_dim, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), scope='RESE'):
        super(ResE, self).__init__(scope)
        H, W, C = in_shape
        min_size = min(H, W)
        ichs = {64: [base_chs * c for c in (1, 2, 4, 4)],
                128: [base_chs * c for c in (1, 2, 4, 8)],
                256: [base_chs * c for c in (1, 2, 4, 8, 8)]}
        ochs = {64: [base_chs * c for c in (2, 4, 4, 8)],
                128: [base_chs * c for c in (2, 4, 8, 16)],
                256: [base_chs * c for c in (2, 4, 8, 8, 16)]}
        factor = {64: base_chs / 32,
                  128: base_chs / 16,
                  256: base_chs / 64}
        self.ichs = ichs[min_size]
        self.ochs = ochs[min_size]

        self.relu = dock.Relu()
        self.flat = dock.Reshape()
        pad = dock.autopad((3, 3), size=(H, W))
        if spec_norm:
            self.conv = dock.SN(dock.Conv(C, self.ichs[0], (3, 3), padding=pad))
        else:
            self.conv = dock.Conv(C, self.ichs[0], (3, 3), padding=pad)

        if norm_fn == CBN:
            norm_fn = BN
        if norm_fn == CIN:
            norm_fn = IN

        for i in range(len(self.ichs)):
            pad = dock.autopad((2, 2), stride=2, size=(H // 2**i, W // 2**i))
            setattr(self, 'pool_%d'%i, dock.AvgPool((2, 2), padding=pad))
            setattr(self, 'blk_%d'%i,
                    ResBlock((H // 2**(i+1), W // 2**(i+1), self.ichs[i]), self.ichs[i]//2, self.ochs[i],
                              norm_fn=norm_fn, attention=attention, spec_norm=spec_norm, w_init=w_init))

        if norm_fn == BN:
            self.nf = dock.BN(self.ochs[-1], dim=2)
        elif norm_fn == IN:
            self.nf = dock.IN(self.ochs[-1], dim=2)
        elif norm_fn == LN:
            self.nf = dock.LN((H // 2**len(self.ichs), W // 2**len(self.ichs), self.ochs[-1]))
        elif norm_fn is None:
            self.nf = dock.Identity()

        hidden_dim = int(H * W * factor[min_size])
        self.mu = dock.Dense(hidden_dim, latent_dim)
        self.log_var = dock.Dense(hidden_dim, latent_dim)
        self.exp = dock.Exp()

    def reparam(self, mu, lv):
        eps = dock.coat(np.random.normal(size=mu.shape).astype(np.float32))
        return mu + self.exp(lv / 2) * eps

    def run(self, x):
        bs = x.shape[0]
        self['input'] = x
        x = self.conv(x)
        x = self.relu(x)

        for i in range(len(self.ichs)):
            pl = getattr(self, 'pool_%d'%i)
            rb = getattr(self, 'blk_%d'%i)
            x = pl(x)
            x = rb(x)

        x = self.nf(x)
        x = self.relu(x)

        self['hidden'] = self.flat(x, (bs, -1))
        mu = self.mu(self['hidden'])
        lv = self.log_var(self['hidden'])
        z = self.reparam(mu, lv)
        return mu, lv, z


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import os
import csv
import h5py
from collections import abc
from multiprocessing import cpu_count
from math import ceil
import torch
from torch.utils.data import Dataset, DataLoader
from torch.utils.data._utils.collate import default_collate
from torch import cuda

from ..law import Constant
from ..kit.utility import ver2num

__all__ = ('Tank', 'Depot', 'load_h5', 'load_csv')


if ver2num(torch.__version__) >= ver2num('1.7.0'):
    is_new_version = True
else:
    is_new_version = False


def load_h5(data_path):
    assert data_path.endswith('h5') or data_path.endswith('hdf5')
    hdf5 = h5py.File(data_path, 'r')
    return hdf5

def load_csv(data_path, data_specf):
    assert data_path.endswith('csv') or data_path.endswith('txt')
    tdata = {}
    with open(data_path, 'r') as csvf:
        csvr = csv.reader(csvf, delimiter=Constant.CHAR_SEP, quotechar=Constant.FIELD_SEP)
        for l, line in enumerate(csvr):
            if l == 0:
                keys = line
                tdata = {k: [] for k in keys}
                continue
            for i, val in enumerate(line):
                if data_specf[keys[i]].startswith('int'):
                    tdata[keys[i]].append(int(val))
                elif data_specf[keys[i]].startswith('float'):
                    tdata[keys[i]].append(float(val))
                else:
                    tdata[keys[i]].append(val)
    return tdata



class Tank(Dataset):
    def __init__(self, *args, **kwargs):
        self._length = self.load(*args, **kwargs)

    def load(self, *args, **kwargs) -> int:
        raise NotImplementedError

    def fetch(self, idx):
        raise NotImplementedError

    def collate(self, batch):
        return default_collate(batch)

    def __getitem__(self, item):
        return self.fetch(item)

    def __len__(self):
        return self._length


class Depot(object):
    def __init__(self, engine):
        self.rank = int(os.environ.get('RANK', -1))
        self.nworld = int(os.environ.get('WORLD_SIZE', 1))
        if hasattr(engine, 'chip'):
            if engine.multi_piston:
                self._chip = torch.device('cuda')
            else:
                self._chip = engine.chip[self.rank]
            self._coater = engine.coat
        else:
            self._chip = None
        self._tanks = {}
        self._batch_size = {}
        self.MPE = {}

    def mount(self, tank, batch_size, shuffle=True, in_same_size=True, nworker=-1, prefetch=0, transmit=False):
        # >| get unique id
        tid = '0t' + hex(id(tank))[2:]
        # >| check argument validity
        if tid in self._tanks.keys():
            raise AttributeError('NEBULAE ERROR ⨷ this tank has already been mounted.')
        if nworker == 0 and prefetch > 0:
            raise ValueError('NEBULAE ERROR ⨷ the number of workers should be more than 1 when prefetcher is enabled.')
        assert nworker != 0, 'NEBULAE ERROR ⨷ the number of workers must be a positive integer.'
        if transmit and (self._chip is None or not cuda.is_available()):
            print('NEBULAE WARNING ◘ you have no GPU to which the data is transmitting.')
            transmit = False
        self._transmit = transmit
        if nworker == 1:
            prefetch = -1
        else:
            if is_new_version:
                prefetch = max(1, prefetch)  # at least be 1 in pytorch
            else:
                print('NEBULAE WARNING ◘ to prefetch on CPU side, please upgrade PyTorch to 1.7.0 or higher.')
                prefetch = -1

        # >| mount dataset
        self._batch_size[tid] = batch_size
        if in_same_size:
            self.MPE[tid] = len(tank) // (batch_size * self.nworld)
        else:
            self.MPE[tid] = ceil(len(tank) / (batch_size * self.nworld))
        nworker = cpu_count() if nworker <= 0 else nworker - 1
        if self.rank >= 0:
            from torch.utils.data import distributed as dist
            sampler = dist.DistributedSampler(tank)
            if prefetch < 0:
                loader = DataLoader(tank, batch_size, shuffle, sampler=sampler, collate_fn=tank.collate,
                                          drop_last=in_same_size, num_workers=nworker)
            else:
                loader = DataLoader(tank, batch_size, shuffle, sampler=sampler, collate_fn=tank.collate,
                                    drop_last=in_same_size, num_workers=nworker, prefetch_factor=prefetch)
        else:
            if prefetch < 0:
                loader = DataLoader(tank, batch_size, shuffle, collate_fn=tank.collate,
                                          drop_last=in_same_size, num_workers=nworker)
            else:
                loader = DataLoader(tank, batch_size, shuffle, collate_fn=tank.collate,
                                    drop_last=in_same_size, num_workers=nworker, prefetch_factor=prefetch)
        # >| prefetch if necessary
        if self._transmit: # no matter how many batches to prefetch on GPU is all the same
            self._tanks[tid] = [tank, loader, loader.__iter__(), 0, cuda.Stream(device=self._chip)]
            self._prefetch(tid)
        else:
            self._tanks[tid] = [tank, loader, loader.__iter__(), 0]
        return tid

    def unmount(self, tid):
        assert tid in self._tanks.keys(), 'NEBULAE ERROR ⨷ this tank is not mounted in the depot.'
        self._tanks.pop(tid)
        self._batch_size.pop(tid)
        self.MPE.pop(tid)

    def _prefetch(self, tid):
        self._fetched_data = self._fetch(tid)
        with cuda.stream(self._tanks[tid][4]):
            # self._fetched_data = self._fetched_data.cuda(device=self.chip, non_blocking=True)
            if isinstance(self._fetched_data, torch.Tensor):
                self._fetched_data = self._coater(self._fetched_data, sync=False)
            elif isinstance(self._fetched_data, abc.Mapping):
                self._fetched_data = {key: self._coater(self._fetched_data[key]) for key in self._fetched_data}
            elif isinstance(self._fetched_data, abc.Sequence):
                self._fetched_data = [self._coater(fd) for fd in self._fetched_data]
            else:
                raise TypeError('NEBULAE ERROR ⨷ batchified data should be a tensor, sequence or dictionary but is %s'
                                % type(self._fetched_data))

    def _fetch(self, tid):
        loader, iterator, counter = self._tanks[tid][1:4]
        if counter == self.MPE[tid]:
            iterator = loader.__iter__()
            self._tanks[tid][2:4] = iterator, 1
        else:
            self._tanks[tid][3] += 1
        return iterator.__next__()

    def _next(self, tid):
        if self._transmit:
            cuda.current_stream().wait_stream(self._tanks[tid][4])
            fetched_data = self._fetched_data
            self._prefetch(tid)
            return fetched_data
        else:
            return self._fetch(tid)

    def next(self, tid=None):
        if tid is None:
            ret = {}
            for k in self._tanks.keys():
                ret[k] = self._next(k)
            return ret
        else:
            return self._next(tid)


#!/usr/bin/env python
'''
vae
Created by Seria at 2022/11/3 4:06 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock, Craft
from .architect import MLPE, MLPD





class VAE(Craft):
    def __init__(self, in_shape, hidden_dim, latent_dim, scope='VAE'):
        super(VAE, self).__init__(scope)
        self.E = MLPE(in_shape, hidden_dim, latent_dim)
        self.D = MLPD(in_shape, hidden_dim, latent_dim)

    def run(self, x):
        mu, lv, z = self.E(x)
        y = self.D(z)
        return mu, lv, y